(* Les graphes seront représentés sous forme de listes d'adjacence *)
type sommet = int ;;
type graphe = ( sommet list ) array ;;
type couleur = AUCUNE | COULEUR of int ;;

let est_2_coloriable ( g : graphe ) : bool =
  let inverse = function
      | AUCUNE -> AUCUNE
      | COULEUR(i) -> COULEUR( (i + 1) mod 2 ) in
  let couleur = Array.make ( Array.length g ) AUCUNE in
  let a_visiter = Queue.create () in
  let rec parcours_composante ( a_visiter : sommet Queue.t ) =
    if Queue.is_empty a_visiter then true
    else
    begin
      let s = Queue.pop a_visiter in
      let rec parcours_voisins ( voisins : sommet list ) : bool =
        match voisins with
        | [] -> true
        | v :: voisins when couleur.(v) = couleur.(s) -> false
        | v :: voisins when couleur.(v) = AUCUNE ->
          begin
            couleur.(v) <- (inverse couleur.(s)) ;
            Queue.push v a_visiter ;
            parcours_voisins voisins
          end
        | v :: voisins -> parcours_voisins voisins in
      if parcours_voisins g.(s) then parcours_composante a_visiter
      else false
    end in
  let rec parcours ( i : sommet ) : bool =
    if i = Array.length g then true
    else
      if couleur.(i) = AUCUNE then
      begin
        couleur.(i) <- COULEUR(0);
        Queue.push i a_visiter;
        if parcours_composante a_visiter then parcours ( i + 1 )
        else false
      end
      else parcours ( i + 1 ) in
  parcours 0
;;

let coloration_glouton ( g : graphe ) : couleur array =
  let couleur = Array.make ( Array.length g ) AUCUNE in
  let visite_prevue = Array.make ( Array.length g ) false in
  let a_visiter = Queue.create () in
  let rec parcours_composante
            ( a_visiter : sommet Queue.t ) ( nb_couleur : int ) : unit =
    if Queue.is_empty a_visiter then ()
    else
    begin
      let couleur_utilisee = Array.make nb_couleur false in
      let s = Queue.pop a_visiter in
      let rec parcours_voisins ( voisins : sommet list ) : unit =
        match voisins with
        | [] -> ()
        | v :: voisins -> match couleur.(v) with
          | AUCUNE -> ( if not visite_prevue.(v) then
                          ( Queue.push v a_visiter ; visite_prevue.(v) <- true );
                            parcours_voisins voisins )
          | COULEUR(k) -> ( couleur_utilisee.(k) <- true;
                            parcours_voisins voisins ) in
      parcours_voisins g.(s);
      let couleur_disponible = ref None in
      let i = ref 0 in
      while !couleur_disponible = None && !i < nb_couleur do
        if not (couleur_utilisee.(!i)) then
          couleur_disponible := Some(!i)
        else incr i
      done;
      match !couleur_disponible with
      | None -> ( couleur.(s) <- COULEUR(nb_couleur);
                  parcours_composante a_visiter ( nb_couleur + 1 ) )
      | Some(x) -> ( couleur.(s) <- COULEUR(x) ;
                     parcours_composante a_visiter nb_couleur )
    end in
  let rec parcours ( i : sommet ) : unit =
    if i = Array.length g then ()
    else
    begin
      if couleur.(i) = AUCUNE then
      begin
        Queue.push i a_visiter;
        visite_prevue.(i) <- true;
        parcours_composante a_visiter 1;
      end;
      parcours ( i + 1 )
    end in
  parcours 0;
  couleur
;;

let coloration_welsh_powell ( g : graphe ) : couleur array =
  let nb_sommet = Array.length g in
  let couleur = Array.make nb_sommet AUCUNE in
  let degre = Array.make nb_sommet 0 in
  let tri_sommet = Array.make nb_sommet 0 in
  let nb_couleur = ref 1 in
  for i = 0 to nb_sommet - 1 do
    degre.(i) <- List.length g.(i) ;
    tri_sommet.(i) <- i
  done;
  Array.stable_sort (fun x y -> compare degre.(x) degre.(y)) tri_sommet;
  for i = 0 to nb_sommet - 1 do
    let couleur_utilisee = Array.make !nb_couleur false in
    let s = tri_sommet.(i) in
    let rec parcours_voisins ( voisins : sommet list ) : unit =
      match voisins with
      | [] -> ()
      | v :: voisins -> match couleur.(v) with
        | AUCUNE -> parcours_voisins voisins
        | COULEUR(k) -> ( couleur_utilisee.(k) <- true;
                          parcours_voisins voisins ) in
    parcours_voisins g.(s);
    let couleur_disponible = ref None in
    let i = ref 0 in
    while !couleur_disponible = None && !i < !nb_couleur do
      if not (couleur_utilisee.(!i)) then
        couleur_disponible := Some(!i)
      else incr i
    done;
    match !couleur_disponible with
    | None -> ( couleur.(s) <- COULEUR(!nb_couleur) ; incr nb_couleur )
    | Some(x) -> couleur.(s) <- COULEUR(x) ;
    parcours_voisins g.(s);
  done ;
  couleur
;;

let est_3coloriable ( g : graphe ) : bool =
  let nb_sommet = Array.length g in
  let visite_prevue = Array.make nb_sommet false in
  let rec parcours_composante
            ( a_visiter : sommet Queue.t )
            ( couleur : couleur array ) : bool =
    if Queue.is_empty a_visiter then true
    else
    begin
      let couleur_utilisee = Array.make 3 false in
      let s = Queue.pop a_visiter in
      visite_prevue.(s) <- true;
      let rec parcours_voisins ( voisins : sommet list ) : unit =
        match voisins with
        | [] -> ()
        | v :: voisins -> match couleur.(v) with
          | AUCUNE -> ( if not visite_prevue.(v) then
                          ( Queue.push v a_visiter ; visite_prevue.(v) <- true );
                            parcours_voisins voisins )
          | COULEUR(k) -> ( couleur_utilisee.(k) <- true;
                            parcours_voisins voisins ) in
      parcours_voisins g.(s);
      let est_3coloriable = ref false in
      let couleur_testee = ref 0 in
      while !couleur_testee < 3 && not !est_3coloriable do
        if not couleur_utilisee.(!couleur_testee) then
        begin
          let copy_couleur = Array.copy couleur in
          let copy_a_visiter = Queue.copy a_visiter in
          copy_couleur.(s) <- COULEUR(!couleur_testee);
          est_3coloriable := parcours_composante copy_a_visiter copy_couleur;
        end;
        incr couleur_testee
      done;
      !est_3coloriable
    end in
  let rec parcours ( i : sommet ) : bool =
    if i = Array.length g then true
    else
    begin
      if not visite_prevue.(i) then
      begin
        let a_visiter = Queue.create () in
        Queue.push i a_visiter;
        visite_prevue.(i) <- true;
        if parcours_composante a_visiter (Array.make nb_sommet AUCUNE)
        then parcours ( i + 1 )
        else false
      end
      else parcours ( i + 1 )
    end in
  parcours 0;
;;

let nombre_chromatique ( g : graphe ) : int =
  let nb_sommet = Array.length g in
  let visite_prevue = Array.make nb_sommet false in
  let rec parcours_composante
            ( a_visiter : sommet Queue.t )
            ( nb_couleur : int )
            ( couleur : couleur array ) : int =
    if Queue.is_empty a_visiter then nb_couleur
    else
    begin
      let couleur_utilisee = Array.make nb_couleur false in
      let s = Queue.pop a_visiter in
      let rec parcours_voisins ( voisins : sommet list ) : unit =
        match voisins with
        | [] -> ()
        | v :: voisins -> match couleur.(v) with
          | AUCUNE -> ( if not visite_prevue.(v) then
                          ( Queue.push v a_visiter ; visite_prevue.(v) <- true );
                            parcours_voisins voisins )
          | COULEUR(k) -> ( couleur_utilisee.(k) <- true;
                            parcours_voisins voisins ) in
      parcours_voisins g.(s);
      let nb_chromatique = ref ( (Array.length g) + 1 ) in
      for i = 0 to nb_couleur - 1 do
        if not couleur_utilisee.(i) then
        begin
          let copy_couleur = Array.copy couleur in
          let copy_a_visiter = Queue.copy a_visiter in
          copy_couleur.(s) <- COULEUR(i);
          nb_chromatique :=
            min
              !nb_chromatique
              ( parcours_composante copy_a_visiter nb_couleur copy_couleur ) ;
        end
      done;
      let copy_couleur = Array.copy couleur in
      let copy_a_visiter = Queue.copy a_visiter in
      copy_couleur.(s) <- COULEUR(nb_couleur);
      nb_chromatique :=
        min
          !nb_chromatique
          ( parcours_composante
              copy_a_visiter ( nb_couleur + 1 ) copy_couleur ) ;
      !nb_chromatique
    end in
  let rec parcours ( i : sommet ) ( nb_chromatique : int ): int =
    if i = Array.length g then nb_chromatique
    else
    begin
      if not visite_prevue.(i) then
      begin
        let a_visiter = Queue.create () in
        Queue.push i a_visiter;
        visite_prevue.(i) <- true;
        let nb_chromatique_composante =
          parcours_composante a_visiter 1 (Array.make nb_sommet AUCUNE) in
        parcours ( i + 1 ) ( max nb_chromatique nb_chromatique_composante )
      end
      else
        parcours ( i + 1 ) nb_chromatique
    end in
  parcours 0 1;
;;
